# -*- coding: utf-8 -*-
"""

LETKF plot suite

Plots hourly output data from MONARCH LETKF, compute misfits,
innovation statistics, model fields, analysis increments
and ensemble standard deviation

TODO:
    - [ ] use experiment namelist to set script parameters
    - [ ] generalize instrument type (only VIIRS DOD currently)
"""

import os
import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import matplotlib.colors as mc
import cartopy.crs as ccrs
import obsdb
from datetime import datetime as dt
from datetime import timedelta as td
from scipy.interpolate import griddata
import copy
import subprocess

def plot_2Dmap(field,lats,lons,**kwargs):
    
    filename = 'fig.png'
    # colormap = 'viridis'
    colormap = 'YlOrBr'
    if 'filename' in kwargs: filename = kwargs['filename']
    if 'colormap' in kwargs: colormap = kwargs['colormap']
    # if not os.path.exists(os.path.split(filename)[0]):
    #     os.makedirs(os.path.split(filename)[0])
    
    fig = plt.figure(dpi=300)

    # Setup a global EckertIII map with faint coastlines.
    ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
    # ax.set_global()
    if 'plotregion' in kwargs:
        plotregion = kwargs['plotregion']
        ax.set_extent([plotregion['LonMin'], plotregion['LonMax'], 
                       plotregion['LatMin'], plotregion['LatMax']], crs=ccrs.PlateCarree())
        
    ax.coastlines('110m', alpha=0.1)
    
    vmin, vmax = field.min(), field.max()
    if 'range' in kwargs and kwargs['range'] is not None:
        if "min" in kwargs['range']: vmin = kwargs['range']["min"]
        if "max" in kwargs['range']: vmax = kwargs['range']["max"]
        
    levels=np.linspace(vmin,vmax,11)
    norm = mc.BoundaryNorm(levels, 256)
    filled_c = ax.contourf(lons, lats, field, 
                           transform=ccrs.PlateCarree(),
                           cmap=colormap,
                           levels=levels,
                           norm=norm,
                           extend='both')
    
    # And black line contours.
    # line_c = ax.contour(lons, lats, field, levels=filled_c.levels,
    #                 colors=['black'],
    #                 transform=ccrs.PlateCarree())

    # Uncomment to make the line contours invisible.
    # plt.setp(line_c.collections, visible=False)
    
    # Add a colorbar for the filled contour.
    cbar = fig.colorbar(filled_c, orientation='horizontal')

    # Use the line contours to place contour labels.
    # ax.clabel(
    #     line_c,  # Typically best results when labelling line contours.
    #     colors=['black'],
    #     manual=False,  # Automatic placement vs manual placement.
    #     inline=True,  # Cut the line where the label will be placed.
    #     fmt=' {:.0f} '.format,  # Labes as integers, with some extra space.
    # )
    
    if 'title' in kwargs: ax.set_title(kwargs['title'])

    # plt.show()
    plt.savefig(filename)

dastep = td(hours=3)
dawindow = td(hours=24)
isregional = True
# MONARCH grid information (used to rotate back observations lats,lons)
cenlon, cenlat = 20., 35.
# cenlon is tph0d, cenlat is tlm0d
exppath = '/esarchive/exp/monarch/a6dv'
# exppath = '/esarchive/exp/monarch/a657'
mpath = os.path.join(exppath,'regional/3hourly/ensemble-stats')
dpath = os.path.join(exppath,'diags')
mvar = 'od550du'
stype = 'av'
date = '2020060100'
# date = '2012010100'
mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,date,stype))
pltbounds = {'LatMin':-11,'LatMax':70,'LonMin':-62,'LonMax':102}
# pltbounds = {'LatMin':15,'LatMax':55,'LonMin':-10,'LonMax':50}
# pltbounds = {'LatMin':-90,'LatMax':90,'LonMin':-180,'LonMax':180}
maprange = {'od550du':{'max':1.0,'min':0.}}
mapdiffrange = {'od550du':{'max':0.5,'min':-0.5}}
histrange = {'od550du':{'max':1.5,'min':0.}}
histdiffrange = {'od550du':{'max':1.5,'min':-1.5}}
markersize = 1.0
outdir = '/esarchive/exp/monarch/a6dv/plots'

if not os.path.exists(outdir): os.makedirs(outdir)

inidate = dt.strptime(date, '%Y%m%d%H')
dasteps = [inidate+i*dastep for i in range(int(dawindow/dastep)+1)]

obsfile = os.path.join(dpath,'dep.txt')
rejobsfile = os.path.join(dpath,'dep_rej.txt')

# yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
nobs, nused, nrej, y_ts, hofxbg_ts, hofxan_ts, fgdep_ts, andep_ts = [[] for i in range(8)]
# Now loop on output frequency and do some plots
for step in dasteps[1:]:
    istep = int((step-inidate)/dastep)
    print(step)
    print(istep)
    
    # Read and plot observations, departures and hofxbg. 
    # Object are reinitialized to avoid cumulating the obs for the moment
    y = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
    yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
    fgdep = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
    # Read the obs here from the LETFK output (consider appending the obs, it should be automatically done)
    # but take care of the correct time with this slot shit
    y.ReadLETKFAscii(obsfile,inidate,islot=istep,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
    fgdep.ReadLETKFAscii(obsfile,inidate,islot=istep,ReadDA={'ObsType':'dep'},rotated=isregional,cenlon=cenlon,cenlat=cenlat)
    naccepted = yall.ReadLETKFAscii(obsfile,inidate,islot=istep,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
    nrejected = yall.ReadLETKFAscii(rejobsfile,inidate,islot=istep,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
    # Flag rejected obs
    if nrejected > 0: yall.assim_mask[-nrejected:] = 0

    # I can interpolate both hofx_fg and hofx_an. Alternatively the first could be read from file.
    hofxbg = y - fgdep
    # hofxbg = copy.deepcopy(y)
    hofxan = copy.deepcopy(y)
    
    y.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                # 'colormap':'viridis'},
                                'colormap':'YlOrBr'},
                    PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                    outNameSuffix='y',outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    y.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                        outNameSuffix='y',outdir=outdir,subtitle='y '+step.strftime('%Y-%m-%d-%H'))
    
    # FG ensemble average
    stype = 'av'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xfg = f[mvar][istep,0,:,:]
    lat_grid = f['lat'][:]
    lon_grid = f['lon'][:]
    f.close()
    
    # AN ensemble average
    stype = 'av_an'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xan = f[mvar][istep,0,:,:]
    f.close()
    
    stype = 'av_inc'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xinc = f[mvar][istep,0,:,:]
    f.close()
    
    # FG ensemble standard deviation
    stype = 'std'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xfg_std = f[mvar][istep,0,:,:]
    f.close()
    
    # AN ensemble standard deviation
    stype = 'std_an'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xan_std = f[mvar][istep,0,:,:]
    f.close()
    
    # if not xfg.mask.any():
    plot_2Dmap(xfg,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2f Min:%.2f Max:%.2f' % (xfg.mean(),xfg.min(),xfg.max()),
               plotregion=pltbounds,
               filename=os.path.join(outdir,'momarch-aod-%s-fg' % (step.strftime('%Y%m%d%H'))))
    
    plot_2Dmap(xan,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' AN\n Avg:%.2f Min:%.2f Max:%.2f' % (xan.mean(),xan.min(),xan.max()),
               plotregion=pltbounds,
               filename=os.path.join(outdir,'momarch-aod-%s-an' % (step.strftime('%Y%m%d%H'))))
    
    plot_2Dmap(xinc,lat_grid,lon_grid,range=mapdiffrange[mvar],
               title=step.isoformat()+' INC \n Avg:%.2f Min:%.2f Max:%.2f' % (xinc.mean(),xinc.min(),xinc.max()),
               plotregion=pltbounds,
               filename=os.path.join(outdir,'momarch-aod-%s-inc' % (step.strftime('%Y%m%d%H'))),colormap='Spectral_r')
    
    plot_2Dmap(xfg_std,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' FG std\n Avg:%.2f Min:%.2f Max:%.2f' % (xfg_std.mean(),xfg_std.min(),xfg_std.max()),
               plotregion=pltbounds,
               filename=os.path.join(outdir,'momarch-aod-%s-fg-std' % (step.strftime('%Y%m%d%H'))))
    
    plot_2Dmap(xan_std,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' AN std\n Avg:%.2f Min:%.2f Max:%.2f' % (xan_std.mean(),xan_std.min(),xan_std.max()),
               plotregion=pltbounds,
               filename=os.path.join(outdir,'momarch-aod-%s-an-std' % (step.strftime('%Y%m%d%H'))))

    # -------------------------------------------------------------------------
    # Interpolate analysis and first guess data at the observation locations to get the innovation values
    # This is temporary, it should disappear when the LETKF will produce all diags files
    
    # interpolated_data = griddata(
    #     (lon_grid.ravel(), lat_grid.ravel()),  # Known points
    #     xfg.ravel(),  # Known data
    #     np.array((y.lons, y.lats)).T,  # Points to interpolate to
    #     method='linear'  # You can use 'linear', 'nearest', or other methods
    # )

    # hofxbg.aod['550'] = np.ma.masked_invalid(interpolated_data)
    # hofxbg.mask = hofxbg.aod['550'].mask
    # hofxbg.lats.mask = hofxbg.mask
    # hofxbg.lons.mask = hofxbg.mask
    # fgdep = y - hofxbg
    # fgdep.mask = hofxbg.mask
    # fgdep.lats.mask = fgdep.mask
    # fgdep.lons.mask = fgdep.mask
    
    interpolated_data = griddata(
        (lon_grid.ravel(), lat_grid.ravel()),  # Known points
        xan.ravel(),  # Known data
        np.array((y.lons, y.lats)).T,  # Points to interpolate to
        method='linear'  # You can use 'linear', 'nearest', or other methods
    )

    hofxan.aod['550'] = np.ma.masked_invalid(interpolated_data)
    hofxan.mask = hofxan.aod['550'].mask
    hofxan.lats.mask = hofxan.mask
    hofxan.lons.mask = hofxan.mask
    andep = y - hofxan
    andep.mask = hofxan.mask
    andep.lats.mask = andep.mask
    andep.lons.mask = andep.mask
    # -------------------------------------------------------------------------
    
    hofxbg.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                            'colormap':'YlOrBr'},
                    PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                    outNameSuffix='hofxbg',outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    hofxbg.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                        outNameSuffix='hofxbg',outdir=outdir,subtitle='hofxbg '+step.strftime('%Y-%m-%d-%H'))
    
    hofxan.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                            'colormap':'YlOrBr'},
                    PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                    outNameSuffix='hofxan',outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    hofxan.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                        outNameSuffix='hofxan',outdir=outdir,subtitle='hofxan '+step.strftime('%Y-%m-%d-%H'))
    
    fgdep.PlotObsValues(ObsOnMapOpt={'range':mapdiffrange[mvar],'markersize':markersize,
                            'colormap':'Spectral_r'},
                    PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                    outNameSuffix='fgdep',outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    fgdep.PlotObsHisto(vmin=histdiffrange[mvar]['min'],vmax=histdiffrange[mvar]['max'],
                       outNameSuffix='fgdep',outdir=outdir,subtitle='fgdep '+step.strftime('%Y-%m-%d-%H'))
    
    andep.PlotObsValues(ObsOnMapOpt={'range':mapdiffrange[mvar],'markersize':markersize,
                        'colormap':'Spectral_r'},
                PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                outNameSuffix='andep',outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    andep.PlotObsHisto(vmin=histdiffrange[mvar]['min'],vmax=histdiffrange[mvar]['max'],
                       outNameSuffix='andep',outdir=outdir,subtitle='andep '+step.strftime('%Y-%m-%d-%H'))
    
    yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
                               PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                               outdir=outdir,subtitle=step.strftime('%Y-%m-%d-%H'))
    
    # Compute stats and save them for a time series plot (not perfect solution since the obs
    # plots should be done by the class methods, but I haven't yet developed strategy to handle
    # multiple obs objects within the obs class)
    nobs.append(yall.nobs)
    nused.append(naccepted)
    nrej.append(nrejected)
    y_ts.append(y.Stats())
    hofxbg_ts.append(hofxbg.Stats())
    hofxan_ts.append(hofxan.Stats())
    fgdep_ts.append(fgdep.Stats())
    andep_ts.append(andep.Stats())
    # This can undo selections if needed
    # y.ResetSel()

# Plot the time series of observations here
fig = plt.figure(figsize=(8,16),dpi=250)

fig.autofmt_xdate()

ax = fig.add_subplot(311)
ax.set_title("Number of %s observations" % ('VIIRS'))

ax.plot(dasteps[1:],nused,color='royalblue',\
        ls='-',lw=2,marker=None,label='Assimilated',markersize=4.0)
ax.plot(dasteps[1:],nrej,color='black',\
    ls='--',lw=2,marker=None,label='Rejected',markersize=3.0)
# ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))

ax.grid(True)   
ax.set_xlabel('Time (UTC)')
ax.legend(loc=1)

ax = fig.add_subplot(312)
ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
ax.set_ylabel("%s" % ('unitless'))

ax.errorbar(dasteps[1:],[v['550']['avg'] for v in hofxbg_ts],
            yerr=[v['550']['std'] for v in hofxbg_ts],color='royalblue',
        ls=None,marker='o',label='hofx_fg',markersize=4.0,
        capsize=6, elinewidth=1.7, markeredgewidth=1.7)
ax.errorbar(dasteps[1:],[v['550']['avg'] for v in hofxan_ts],
            yerr=[v['550']['std'] for v in hofxan_ts],color='salmon',
        ls=None,marker='o',label='hofx_an',markersize=4.0,
        capsize=4, elinewidth=1.3, markeredgewidth=1.3)    
ax.errorbar(dasteps[1:],[v['550']['avg'] for v in y_ts],
            yerr=[v['550']['std'] for v in y_ts],color='black',
        ls=None,marker='o',label='y',markersize=4.0,
        capsize=2, elinewidth=1., markeredgewidth=1.,alpha=0.7)
        
    
ax.grid(True)   
ax.set_xlabel('Time (UTC)')
ax.legend(loc=1)

ax = fig.add_subplot(313)
ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
ax.set_ylabel("%s" % ('unitless'))

ax.errorbar(dasteps[1:],[v['550']['avg'] for v in fgdep_ts],
            yerr=[v['550']['std'] for v in fgdep_ts],color='royalblue',
        ls=None,marker='o',label='fgdep',markersize=4.0,
        capsize=6, elinewidth=1.7, markeredgewidth=1.7)
ax.errorbar(dasteps[1:],[v['550']['avg'] for v in andep_ts],
            yerr=[v['550']['std'] for v in andep_ts],color='salmon',
        ls=None,marker='o',label='andep',markersize=4.0,
        capsize=4, elinewidth=1.3, markeredgewidth=1.3)
    
ax.grid(True)   
ax.set_xlabel('Time (UTC)')
ax.legend(loc=1)

plt.savefig(os.path.join(outdir,'VIIRS_time_series.png'))
plt.close(fig)

#Make some gifs here
for plttype in ['momarch-aod-*-fg.png','momarch-aod-*-an.png','momarch-aod-*-inc.png',
                'momarch-aod-*-fg-std.png','momarch-aod-*-an-std.png',
                'VIIRS_550_AODMap_*_y.png','VIIRS_550_AODMap_*_fgdep.png','VIIRS_550_AODMap_*_andep.png',
                'VIIRS_550_AODMap_*_hofxbg.png','VIIRS_550_AODMap_*_hofxan.png',
                'VIIRS_AODHisto_*_andep.png','VIIRS_AODHisto_*_fgdep.png','VIIRS_MaskMap_*.png']:
    subprocess.run(['convert','-delay','80','-loop','0',os.path.join(outdir,plttype),
                    os.path.join(outdir,plttype.replace('*','anim').replace('.png','.gif'))])
        
# Here we can plot obs for all day if needed
# yall.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
#                             # 'colormap':'viridis'},
#                             'colormap':'YlOrBr'},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='yall',outdir=outdir)
# yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='assim_mask',outdir=outdir)

